package com.andreani.xtol.node;

import com.andreani.xtol.util.*;
import javax.servlet.*;
import javax.servlet.http.*;
import net.jini.discovery.DiscoveryListener;
import net.jini.discovery.LookupDiscoveryManager;
import net.jini.discovery.LookupDiscovery;
import net.jini.discovery.DiscoveryEvent;
import net.jini.discovery.LookupLocatorDiscovery;
import net.jini.core.discovery.LookupLocator;
import net.jini.discovery.DiscoveryListener;
import net.jini.core.lease.Lease;
import net.jini.lease.LeaseRenewalManager;
import java.rmi.*;
import java.net.*;
import java.io.*;
import java.util.*;
import java.rmi.server.*;
import net.jini.lookup.*;
import net.jini.lookup.ServiceIDListener;
import net.jini.core.entry.*;
import net.jini.core.lookup.*;
import net.jini.lookup.entry.*;
import net.jini.discovery.*;
import org.apache.log4j.Category;
import org.apache.log4j.BasicConfigurator;

/**
 * This class realize the agent that joins the network and offers to it its
 * services. Is is a simple Servlet that uses its init() method to make it
 * visible as service into the network. A developer who wants to load balance
 * accesses to his server and make it fail tolerant must subclass this class
 * and call super.init() and super.doGet() at the beginning of these
 * local methods.
 *
 * @author    Stefano Andreani
 * @version   1.0
 */
public class TolerantServlet extends HttpServlet implements
      ServiceIDListener
{
   protected Category cat;
   JoinManager jman;
   ServerNode thisServer = null;
   SessionSpaceManager spacem;

   public TolerantServlet()
   {
      super();
      cat = Category.getInstance(this.getClass().getName());
      //BasicConfigurator.configure();
      cat.debug("contructor");
      spacem = new SessionSpaceManager();
   }

   /**
    * Registers itself into the lookup service.
    * Always invoke super.init(ServletConfig) as first operation when you override
    * this method.
    *
    */
   public void init(ServletConfig sconf)
   {
      cat.debug("init()");
      try
      {
         super.init(sconf);
      }
      catch (ServletException ex)
      {
            cat.error(ex.getMessage(),ex);
      }
      if (System.getSecurityManager() == null)
      {
         System.setSecurityManager(new RMISecurityManager());
      }
      System.out.println("Registering server with lookup service...");
      try
      {
         try
         {
            String locAddr = InetAddress.getLocalHost().getHostAddress();
            cat.debug("Local address="+locAddr);
            ServletContext sctx = sconf.getServletContext();
            int port = Integer.parseInt(sctx.getInitParameter("port"));
            cat.debug("port="+port);
            double load = Math.random();
            cat.debug("load="+load);
            thisServer = new ServerNode(locAddr, port, load);
         }
         catch (java.net.UnknownHostException ex)
         {
            cat.error(ex.getMessage(),ex);
         }
         Entry[] ev = thisServer.getEntries();
         /*
         Lease lease = new Lease(
         LeaseRenewalManager myLeaseMgr = new LeaseRenewalManager(lease, 5000, null);
         */
         cat.debug("new JoinManager");
         jman = new JoinManager(thisServer, ev, this, null, null);
         System.out.println("Registering this service...: jman="+jman);
         //*/
      }
      catch (Exception e)
      {
         cat.error(e.getMessage(), e);
      }
   }

   public void destroy()
   {
      jman.terminate();
   }

   void snoop(HttpServletRequest request, HttpServletResponse response)
   throws IOException
   {
        cat.debug("snoop()");
        HttpSession session = request.getSession();
        //cat.debug("session.getId()="+session.getId());
        Enumeration ea = request.getParameterNames();
        while (ea.hasMoreElements())
        {
            String name = (String)ea.nextElement();
            //cat.debug("ea.nextElement()="+name);
            Object value = request.getParameter(name);
            //cat.debug(value);
            session.setAttribute(name, value);
            //cat.debug(session.getAttribute(name));
        }
        response.setContentType("text/html");
        PrintWriter out = response.getWriter();
        out.println("Response from IP "+thisServer.serverIP+" Port "+thisServer.serverPort+"<br>");
        out.println("Request URI = "+request.getRequestURI()+"<br>");
        out.println("Query String = "+request.getQueryString()+"<br>");
        out.println("<b>Request Headers:</b><br>");
        Enumeration eh = request.getHeaderNames();
        while (eh.hasMoreElements())
        {
            String name = (String)eh.nextElement();
            String value = request.getHeader(name);
            out.println(name + " = " + value + "<br>");
        }
        out.println("<b>Session attributes:</b><br>");
        Enumeration es = session.getAttributeNames();
        while (es.hasMoreElements())
        {
            String name = (String)es.nextElement();
            String value = session.getAttribute(name).toString();
            out.println(name + " = " + value + "<br>");
        }
        cat.debug("fine snoop()");
   }

   /**
    * Always invoke super.doGet(req,resp) as first operation when you override
    * this method.
    *
    */
   public void doGet(HttpServletRequest req, HttpServletResponse resp)
          throws ServletException, IOException
   {
      cat.debug("doGet()");
      spacem.snoopSpace();
      if (req.getSession(false)==null) createLocalSession(req);
      snoop(req,resp);
      spacem.snoopSpace();
   }

   void createLocalSession(HttpServletRequest req)
   {
      cat.debug("createLocalSession()");
      HttpSession session = req.getSession();
      cat.debug("req.getCookies();");
      Cookie[] cv = req.getCookies();
      if(cv==null) cat.error("No cookie found");
      cat.debug("cv="+cv);
      boolean fail = false;
      String globsessid = null;
      for (int i = 0; i < cv.length; i++)
      {
         cat.debug("i="+i);
         if (cv[i].getName().equals(Global.COOKIE_NAME_SESSIONID))
         {
            cat.debug(cv[i].getName());
            globsessid = cv[i].getValue();
         }
         if (cv[i].getName().equals(Global.COOKIE_FAIL_OVER))
         {
            cat.debug(cv[i].getName());
            fail = true;
         }
      }
      if (fail) rebuildSession(globsessid, req);
      else session.setAttribute(Global.COOKIE_NAME_SESSIONID,globsessid);
   }

   /**
    * This method create a new HttpSession retrieving session data from a
    * SessionSpace. It is usually used if the session server failed.
    *
    */
   public void/*HttpSession*/ rebuildSession(String globsessid, HttpServletRequest req)
   {
      cat.debug("rebuildSession(globsessid="+globsessid+")");
      spacem.snoopSpace();
      SessionEntry entry = spacem.retrieveEntry(globsessid);
      HttpSession httpSession = req.getSession();
      for (Enumeration e = entry.session.keys() ; e.hasMoreElements() ;)
      {
         String key = (String) e.nextElement();
         httpSession.setAttribute(key,entry.session.get(key));
      }
      spacem.setFailStatus(globsessid,false);
      //return httpSession;
   }

   void debugSpace(String globsessid)
   {
      cat.debug("debugSpace()");
      SessionEntry se = spacem.retrieveEntry(globsessid);
      cat.debug(se);
   }

   /**
    * Debbugging purpose
    *
    */
   public void serviceIDNotify(ServiceID id)
   {
      cat.info("Registered with Service ID: " + id);
   }

}
